@@ -40,7 +40,7 @@ from py_tools_ds.convenience.object_oriented import alias_property
 from py_tools_ds.geo.coord_calc import get_corner_coordinates
 from py_tools_ds.geo.coord_grid import snap_bounds_to_pixGrid
 from py_tools_ds.geo.coord_trafo import mapXY2imXY, imXY2mapXY, transform_any_prj, reproject_shapelyGeometry
-from py_tools_ds.geo.projection import prj_equal, WKT2EPSG, EPSG2WKT, isLocal
+from py_tools_ds.geo.projection import prj_equal, WKT2EPSG, EPSG2WKT, isLocal, CRS
 from py_tools_ds.geo.raster.conversion import raster2polygon
 from py_tools_ds.geo.vector.topology \
     import get_footprint_polygon, polyVertices_outside_poly, fill_holes_within_poly
@@ -85,7 +85,6 @@ class GeoArray(object):
         :param progress:        show progress bars (default: True)
         :param q:               quiet mode (default: False)
         """
-
         # TODO implement compatibility to GDAL VRTs
         if not (isinstance(path_or_array, (str, np.ndarray, GeoArray)) or
            issubclass(getattr(path_or_array, '__class__'), GeoArray)):
@@ -226,7 +225,6 @@ class GeoArray(object):
     @property
     def rows(self):
         """Get the number of rows of the associated image array."""
-
         return self.shape[0]
 
     @property
@@ -240,13 +238,11 @@ class GeoArray(object):
     @property
     def bands(self):
         """Get the number of bands of the associated image array."""
-
         return self.shape[2] if len(self.shape) > 2 else 1
 
     @property
     def dtype(self):
         """Get the numpy data type of the associated image array."""
-
         if self._dtype:
             return self._dtype
         elif self.is_inmem:
@@ -258,7 +254,6 @@ class GeoArray(object):
     @property
     def geotransform(self):
         """Get the GDAL GeoTransform of the associated image, e.g., (283500.0, 5.0, 0.0, 4464500.0, 0.0, -5.0)"""
-
         if self._geotransform:
             return self._geotransform
         elif not self.is_inmem:
@@ -283,13 +278,11 @@ class GeoArray(object):
     @property
     def xgsd(self):
         """Get the X resolution in units of the given or detected projection."""
-
         return self.geotransform[1]
 
     @property
     def ygsd(self):
         """Get the Y resolution in units of the given or detected projection."""
-
         return abs(self.geotransform[5])
 
     @property
@@ -326,8 +319,8 @@ class GeoArray(object):
 
     @property
     def epsg(self):
+        # type: (None) -> int
         """Get the EPSG code of the projection of the GeoArray."""
-
         return WKT2EPSG(self.projection)
 
     @epsg.setter
@@ -571,7 +564,16 @@ class GeoArray(object):
         elif isinstance(given, (tuple, list)):
             # handle requests like geoArr[[1,2],[3,4]  -> not implemented in from_path if array is not in mem
             types = [type(i) for i in given]
+
             if list in types or tuple in types:
+
+                # avoid that the whole cube is read if only data from a single band is requested
+                if not self.is_inmem \
+                   and len(given) == 3 \
+                   and isinstance(given[2], (int, float, np.integer, np.floating)):
+                    band_subset = GeoArray(self.filePath)[:, :, given[2]]
+                    return band_subset[given[:2]]
+
                 self.to_mem()
 
             if len(given) == 3:
@@ -782,10 +784,9 @@ class GeoArray(object):
             # for some reason GDAL reads arbitrary geotransforms as (0, 1, 0, 0, 0, 1) instead of (0, 1, 0, 0, 0, -1)
             self._geotransform[5] = -abs(self._geotransform[5])  # => force ygsd to be negative
 
-            # temp conversion to EPSG needed because GDAL seems to modify WKT string when writing file to disk
-            # (e.g. using gdal_merge) -> conversion to EPSG and back undos that
+            # consequently use WKT1 strings here as GDAL always exports transformation results as WKT1
             wkt = ds.GetProjection()
-            self._projection = EPSG2WKT(WKT2EPSG(wkt)) if not isLocal(wkt) else ''
+            self._projection = CRS(wkt).to_wkt(version="WKT1_GDAL") if not isLocal(wkt) else ''
 
             if 'nodata' not in self._initParams or self._initParams['nodata'] is None:
                 band = ds.GetRasterBand(1)
@@ -818,7 +819,6 @@ class GeoArray(object):
         :param getitem_params:  <list> a list of slices in the form [row_slice, col_slice, band_slice]
         :return out_arr:        <np.ndarray> the output array
         """
-
         ds = gdal.Open(path)
         if not ds:
             raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
@@ -1022,16 +1022,34 @@ class GeoArray(object):
             # ENVI #
             ########
             if fmt == 'ENVI':
-                ds_out.SetMetadata(envi_metadict, 'ENVI')
+                for bidx in range(self.bands):
+                    band = ds_out.GetRasterBand(bidx + 1)
 
-                if 'band_names' in envi_metadict:
-                    for bidx in range(self.bands):
-                        band = ds_out.GetRasterBand(bidx + 1)
+                    if 'band_names' in envi_metadict:
                         bandname = self.metadata.band_meta['band_names'][bidx].strip()
                         band.SetDescription(bandname)
-
                         assert band.GetDescription() == bandname
-                        del band
+
+                    # NOTE: band-specific nodata values are not supported by the ENVI header format
+                    if 'nodata' in envi_metadict:
+                        nodataVal = self.metadata.band_meta['nodata'][bidx]
+                        band.SetNoDataValue(nodataVal)
+                        assert band.GetNoDataValue() == nodataVal
+
+                    del band
+
+                # avoid that band names are written to global meta
+                if 'band_names' in envi_metadict:
+                    del envi_metadict['band_names']
+
+                if 'nodata' in envi_metadict:
+                    del envi_metadict['nodata']
+
+                # set data ignore value in case self.metadata.band_meta contains a unique nodata value
+                if 'nodata' in self.metadata.band_meta and len(set(self.metadata.band_meta['nodata'])) == 1:
+                    envi_metadict['data ignore value'] = str(self.metadata.band_meta['nodata'][0])
+
+                ds_out.SetMetadata(envi_metadict, 'ENVI')
 
                 if 'description' in envi_metadict:
                     ds_out.SetDescription(envi_metadict['description'])
@@ -1052,12 +1070,22 @@ class GeoArray(object):
 
                 for bidx in range(self.bands):
                     band = ds_out.GetRasterBand(bidx + 1)
-                    bandmeta = bandmeta_dict[bidx]
-                    # meta2write = dict((k, repr(v)) for k, v in self.metadata.band_meta.items() if v is not np.nan)
-                    band.SetMetadata(bandmeta)
+                    bandmeta = bandmeta_dict[bidx].copy()
 
-                    if 'band_names' in envi_metadict:
+                    # filter global metadata out
+                    bandmeta = {k: v for k, v in bandmeta.items() if k not in self.metadata.global_meta}
+                    # meta2write = dict((k, repr(v)) for k, v in self.metadata.band_meta.items() if v is not np.nan)
+
+                    if 'band_names' in bandmeta:
                         band.SetDescription(self.metadata.band_meta['band_names'][bidx].strip())
+                        del bandmeta['band_names']
+
+                    if 'nodata' in bandmeta:
+                        band.SetNoDataValue(self.metadata.band_meta['nodata'][bidx])
+                        del bandmeta['nodata']
+
+                    if bandmeta:
+                        band.SetMetadata(bandmeta)
 
                     band.FlushCache()
                     del band
@@ -1264,12 +1292,13 @@ class GeoArray(object):
                 ax = plt.gca()
 
             rows, cols = image2plot.shape[:2]
-            im = ax.imshow(image2plot.astype(np.int),
+            im = ax.imshow(image2plot.astype(int),
                            palette,
                            interpolation=interpolation,
                            extent=(0, cols, rows, 0),
                            vmin=vmin,
-                           vmax=vmax, )  # compressed excludes nodata values
+                           vmax=vmax
+                           )  # compressed excludes nodata values
             plt.show()
 
             return im
@@ -1306,10 +1335,14 @@ class GeoArray(object):
         assert self.projection, "A projection is needed for a map visualization. Got '%s'." % self.projection
 
         # get image to plot
+        # (reproject to LonLat as workaround in case self.epsg is None because cartopy relies on an existing EPSG code)
         nodataVal = nodataVal if nodataVal is not None else self.nodata
-        image2plot, gt, prj = self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap,
-                                                        boundsMapPrj=boundsMapPrj, res_factor=res_factor,
-                                                        nodataVal=nodataVal)
+        gA2plot = GeoArray(*self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap,
+                                                      boundsMapPrj=boundsMapPrj, res_factor=res_factor,
+                                                      nodataVal=nodataVal,
+                                                      out_prj=None if self.epsg is not None else 4326),
+                           nodata=nodataVal)
+        image2plot = gA2plot[:]
 
         # create map
         def get_cartopy_crs_from_epsg(epsg_code):
@@ -1321,19 +1354,21 @@ class GeoArray(object):
                 else:
                     raise NotImplementedError('The show_map() method currently does not support the given projection.')
 
-        crs_in = get_cartopy_crs_from_epsg(self.epsg)
-        crs_out = get_cartopy_crs_from_epsg(out_epsg if out_epsg is not None else self.epsg)
+        crs_in = get_cartopy_crs_from_epsg(gA2plot.epsg)
+        crs_out = get_cartopy_crs_from_epsg(out_epsg if out_epsg is not None else gA2plot.epsg)
 
         fig = plt.figure(figsize=figsize)
         ax = plt.axes(projection=crs_out)
 
-        ax.set_extent(self.box.boundsMap, crs=crs_in)
+        ax.set_extent(gA2plot.box.boundsMap, crs=crs_in)
 
-        palette, vmin, vmax = self._get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal)
+        palette, vmin, vmax = gA2plot._get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal)
         if nodataVal is not None and np.std(image2plot) != 0:  # do not show nodata
             image2plot = np.ma.masked_equal(image2plot, nodataVal)
         ax.imshow(image2plot, cmap=palette, interpolation=interpolation, vmin=vmin, vmax=vmax,
-                  origin='upper', transform=crs_in, extent=list(self.box.boundsMap))
+                  origin='upper', transform=crs_in,
+                  extent=list(gA2plot.box.boundsMap)
+                  )
 
         # show grid lines
         try:
@@ -1366,7 +1401,7 @@ class GeoArray(object):
         import folium
         import geojson
 
-        lonlatPoly = reproject_shapelyGeometry(self.footprint_poly, self.epsg, 4326)
+        lonlatPoly = reproject_shapelyGeometry(self.footprint_poly, self.prj, 4326)
 
         m = folium.Map(location=tuple(np.array(lonlatPoly.centroid.coords.xy).flatten())[::-1])
         gjs = geojson.Feature(geometry=lonlatPoly, properties={})
